@@ -9,7 +9,7 @@
  *
  **************************************************************************
  *
- * Copyright (C) 1995,1996 Gero Kuhlmann <gero@gkminix.han.de>
+ * Copyright (C) 1995-1998 Gero Kuhlmann <gero@gkminix.han.de>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -35,5 +35,5 @@
  */
 
 /* Initialization */
-extern void init_minidos(void far *progadr, void far *dataend);
+extern void init_minidos __P((unsigned long progadr));
 
