@@ -1,10 +1,8 @@
 /*
- * mknbi.c  -  MaKe NetBoot Image
+ * mknbi.c  -  MaKe NetBoot Image for Linux
  *
- * Version 1.5 <Sun Oct 13 10:11:58 MET DST 1996>
- *
- * Copyright (C) 1995 Gero Kuhlmann   <gero@gkminix.han.de>
- * Copyright (C) 1996 Gero Kuhlmann   <gero@gkminix.han.de>
+ * Copyright (C) 1995,1998 Gero Kuhlmann   <gero@gkminix.han.de>
+ * Copyright (C) 1996,1997 Gero Kuhlmann   <gero@gkminix.han.de>
  *                and Markus Gutschke <gutschk@math.uni-muenster.de>
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -22,10 +20,13 @@
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#include "../common.h"
+#include "common.h"
+#ifdef HAVE_INET
 #include <netdb.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#endif
+#include "nblib.h"
 #include "mknbi.h"
 
 #ifndef _MKNBI_H_LINUX_
@@ -33,123 +34,182 @@
 #endif
 
 
-char *progname;
-char nfsdir[MAXPATHLEN];	/* Directory to mount as root via NFS */
-char rootdev[MAXPATHLEN];	/* Name of root device */
-char kname[MAXPATHLEN];		/* Name of kernel image file */
-char outname[MAXPATHLEN];	/* Name of output file */
-char rdname[MAXPATHLEN];	/* Name of ramdisk image file */
-int kimage;			/* File handle for kernel image */
-int outfile;			/* File handle for output file */
-int rdimage = -1;		/* File handle for ramdisk image */
-int rdlocation = -1;		/* Memory location of ramdisk image */
-enum {
-	rd_auto = RD_AUTO,	/* Ramdisk load mode */
-	rd_eom = RD_EOM,
-	rd_fixed = RD_FIXED
-} rdmode = rd_auto;
-int verbose = 0;		/* Verbosity level; currently just 0, 1 or 2 */
+static char *batchname = NULL;		/* name of system to batch process */
+static char *outname = NULL;		/* Name of output file */
+static char *rdname = NULL;		/* Name of ramdisk image file */
+static char *nfsdir = NULL;		/* Directory to mount as root via NFS */
+static char *rootdev = NULL;		/* Name of root device */
+static char *kname = NULL;		/* Name of kernel image file */
+static char *append = NULL;		/* String to append to end of command line */
+static char *addrs = NULL;		/* String containing various addresses */
+static char *rdmodebuf = NULL;		/* Buffer for ramdisk mode string */
+static int rdmode = RD_AUTO;		/* Ramdisk load mode */
+static int cmdl_msize = CMDLMSIZE;	/* Maximum size of command line */
+static int kimage;			/* File handle for kernel image */
+static int outfile;			/* File handle for output file */
+static int rdimage = -1;		/* File handle for ramdisk image */
+static int rdlocation = -1;		/* Memory location of ramdisk image */
 
-int cur_rec_num = -1;		/* Number of current load record */
-struct load_header header;	/* Load header */
-struct load_record *cur_rec;	/* Pointer to current load record */
+static int cur_rec_num = -1;		/* Number of current load record */
+static struct load_header header;	/* Load header */
+static struct load_record *cur_rec;	/* Pointer to current load record */
 
-unsigned char copyrec_buf[SECTSIZE];
+static unsigned char copyrec_buf[SECTSIZE];
+
+
+
+/*
+ * Command line options and arguments
+ */
+static struct cmdopt opts[] = {
+	{ "batch-sys", 'b', strval, {(char **)&batchname}, NULL,
+	  "name of system to process", "SYSTEM"				},
+	{ "outfile", 'o', strval, {(char **)&outname}, NULL,
+	  "name of boot image output file", "FILE"			},
+	{ "ramdisk-image", 'r', strval, {(char **)&rdname}, NULL,
+	  "ramdisk image source file or directory", "FILE|DIR"		},
+	{ "append", 'a', strval, {(char **)&append}, NULL,
+	  "append string to end of kernel command line", "STRING"	},
+	{ "root-dir", 'd', strval, {(char **)&nfsdir}, NULL,
+	  "NFS mounted root dir:\n"
+	  "  STRING syntax: rom|ram|kernel|<dir>", "STRING"				},
+	{ "ip-addrs", 'i', strval, {(char **)&addrs}, NULL,
+	  "IP addresses for Linux kernel:\n"
+	  "  STRING syntax: rom|kernel"
+#ifdef HAVE_INET
+	  "|<IP-addrs>"
+#endif
+							, "STRING"	},
+	{ "kernel", 'k', strval, {(char **)&kname}, NULL,
+	  "path to Linux kernel image", "FILE"				},
+	{ "ramdisk-mode", 'l', strval, {(char **)&rdmodebuf}, NULL,
+	  "loading mode for ramdisk:\n"
+	  "  STRING syntax: auto|eom|<fixed address>", "STRING"		},
+	{ "cmdl-size", 's', intval, {(char **)&cmdl_msize}, NULL,
+	  "maximum command line size", "SIZE"				},
+	{ "kernel", 0, nonopt, {(char **)&kname}, NULL,
+	  "Linux kernel image (if not given as option)", NULL		},
+	{ "outfile", 0, nonopt, {(char **)&outname}, NULL,
+	  "output file (if not given as option)", NULL			},
+	{ NULL, 0, noval, {NULL}, NULL, NULL, NULL			}
+};
+
+
+
+/*
+ * Parameters in each section of database file
+ */
+static struct paramdef dbparams[] = {
+  { "outfile",	 	par_string,	NULL,	{&outname}},
+  { "kernel",		par_string,	NULL,	{&kname}},
+  { "ramdisk-image", 	par_string,	NULL,	{&rdname}},
+  { "ramdisk-mode",	par_string,	NULL,	{&rdmodebuf}},
+  { "append",		par_string,	NULL,	{&append}},
+  { "root-dir",		par_string,	NULL,	{&nfsdir}},
+  { "ip-addrs",		par_string,	NULL,	{&addrs}},
+  { "cmdl-size",	par_int,	NULL,	{(char **)&cmdl_msize}},
+  { NULL,	 	par_null,	NULL,	{NULL}}
+};
 
 
 
 /*
  * Write a buffer into the output file and update the load record
  */
-static void putrec(int recnum, char *src, int size)
+static void putrec(recnum, src, size)
+int recnum;
+char *src;
+int size;
 {
   unsigned long l;
-  long isize;
-  char *buf;
+  size_t isize;
+  unsigned char *buf;
 
   if (cur_rec_num != recnum) {
     fprintf(stderr, "%s: Internal error; image chunks mis-ordered!\n",
 	    progname);
-    exit(EXIT_FAILURE);
+    exit(EXIT_LINUX_MISORDER);
   }
   isize = ((size / (SECTSIZE + 1)) + 1) * SECTSIZE;
-  if ((buf = malloc(isize)) == NULL) {
-	perror(progname);
-	exit(EXIT_FAILURE);
-  }
-  memset(buf, isize, 0);
+  buf = (unsigned char *)nbmalloc(isize);
   memcpy(buf, src, size);
-  if (write(outfile, buf, isize) != isize) {
-	perror(outname);
-	exit(EXIT_FAILURE);
-  }
+  (void)nbwrite(buf, isize, outfile);
   free(buf);
   l = get_long(cur_rec->ilength) + isize;
-  cur_rec->ilength.low  = htot(low_word(l));
-  cur_rec->ilength.high = htot(high_word(l));
+  assign(cur_rec->ilength.low, htot(low_word(l)));
+  assign(cur_rec->ilength.high, htot(high_word(l)));
   l = get_long(cur_rec->mlength) + isize;
-  cur_rec->mlength.low  = htot(low_word(l));
-  cur_rec->mlength.high = htot(high_word(l));
+  assign(cur_rec->mlength.low, htot(low_word(l)));
+  assign(cur_rec->mlength.high, htot(high_word(l)));
 }
 
 
+
 /*
  * Copy a certain number of bytes from the kernel image file into the
  * boot file
  */
-static int copyrec(int recnum, int remaining, int image, char *name)
+static int copyrec(recnum, remaining, image, name)
+int recnum;
+int remaining;
+int image;
+char *name;
 {
-  int bytes_read, size, i = 1;
+  unsigned int size, bytes_read;
+  int i = 1;
 
   if (cur_rec_num != recnum) {
-    fprintf(stderr, "%s: Internal error; image chunks mis-ordered!\n",
-	    progname);
-    exit(EXIT_FAILURE);
+	fprintf(stderr, "%s: Internal error; image chunks mis-ordered!\n",
+								progname);
+	exit(EXIT_LINUX_MISORDER);
   }
   for (bytes_read = 0; remaining > 0 && i > 0;) {
 	size = (remaining > SECTSIZE) ? SECTSIZE : remaining;
-	if ((i = read(image, copyrec_buf, size)) < 0) {
-		perror(name);
-		exit(EXIT_FAILURE);
-	}
-	if (i) {
-		putrec(recnum, copyrec_buf, i);
+	if ((i = nbread(copyrec_buf, size, image))) {
+		putrec(recnum, (char *)copyrec_buf, i);
 		bytes_read += i;
 		remaining -= i;
 	}
   }
-  return (bytes_read);
+  return(bytes_read);
 }
 
 
+
 /*
  * Initialize a load record
  */
-static void initrec(int recnum, int segment, int flags, int vendor_size)
+static void initrec(recnum, segment, flags, vendor_size)
+int recnum;
+int segment;
+int flags;
+int vendor_size;
 {
   if (++cur_rec_num != recnum) {
 	fprintf(stderr, "%s: Internal error; image chunks mis-ordered!\n",
 		progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_MISORDER);
   }
 
   if (cur_rec_num > 0)
-	(unsigned char *)cur_rec += ((cur_rec->rlength << 2) & 0x3c) +
-					((cur_rec->rlength >> 2) & 0x3c);
+	cur_rec = (struct load_record *)((unsigned char *)cur_rec +
+					((cur_rec->rlength << 2) & 0x3c) +
+					((cur_rec->rlength >> 2) & 0x3c));
   cur_rec->rlength      = (sizeof(struct load_record) -
 		           sizeof(union vendor_data)) >> 2;
   cur_rec->rlength     |= ((vendor_size + 3) & 0x3c) << 2;
   cur_rec->rtag1        = recnum + VENDOR_OFF;
   cur_rec->rflags       = flags;
-  cur_rec->address.low  = htot(low_word((unsigned long) segment << 4));
-  cur_rec->address.high = htot(high_word((unsigned long) segment << 4));
+  assign(cur_rec->address.low, htot(low_word((unsigned long) segment << 4)));
+  assign(cur_rec->address.high, htot(high_word((unsigned long) segment << 4)));
 }
 
 
+
 /*
  * Process kernel image file
  */
-static void do_kernel(void)
+static void do_kernel()
 {
   int flags, setup_size, bigkernel;
   struct setup_header *setup_header;
@@ -159,42 +219,43 @@ static void do_kernel(void)
   initrec(INITNUM, DEF_INITSEG, 0, 0);
   if (copyrec(INITNUM, INITLSIZE, kimage, kname) != INITLSIZE) {
 	fprintf(stderr, "%s: Unexpected end of kernel image file\n", progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_KERNEOF);
   }
-  if (*((__u16 *)&(((__u8 *)copyrec_buf)[BOOT_SIG_OFF])) != htot(BOOT_SIGNATURE)){
+  if (getval(*((__u16 *)&(((__u8 *)copyrec_buf)[BOOT_SIG_OFF]))) !=
+							htot(BOOT_SIGNATURE)){
 	fprintf(stderr, "%s: Could not find magic number in kernel image\n",
 		progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_INVKERN);
   }
-  cur_rec->mlength.low  = htot(low_word(INITMSIZE));
-  cur_rec->mlength.high = htot(high_word(INITMSIZE));
+  assign(cur_rec->mlength.low, htot(low_word(INITMSIZE)));
+  assign(cur_rec->mlength.high, htot(high_word(INITMSIZE)));
 
   /* Determine size of setup code */
   setup_size = (int)(((__u8 *)copyrec_buf)[SETUP_SIZE_OFF]) * SECTSIZE;
   if (setup_size > SETUPLSIZE) {
 	fprintf(stderr, "%s: Kernel setup code too large\n", progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_INVSETUP);
   }
 
   /* Read setup header and check if this is a "big" kernel image */
   initrec(SETUPNUM, DEF_SETUPSEG, 0, 0);
   if (copyrec(SETUPNUM, SECTSIZE, kimage, kname) != SECTSIZE) {
 	fprintf(stderr, "%s: Unexpected end of kernel image file\n", progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_KERNEOF);
   }
   setup_header = (struct setup_header *)copyrec_buf;
   bigkernel = !memcmp(setup_header->magic, SETUP_MAGIC, strlen(SETUP_MAGIC)) &&
-               (ttoh(setup_header->version) >= SETUP_VERSION) &&
+               (ttoh(getval(setup_header->version)) >= SETUP_VERSION) &&
                (setup_header->loadflags & SETUP_HIGH);
 
   /* Process the setup code */
   if ((setup_size -= SECTSIZE) > 0 &&
       copyrec(SETUPNUM, setup_size, kimage, kname) != setup_size) {
 	fprintf(stderr, "%s: Unexpected end of kernel image file\n", progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_KERNEOF);
   }
-  cur_rec->mlength.low  = htot(low_word(SETUPMSIZE));
-  cur_rec->mlength.high = htot(high_word(SETUPMSIZE));
+  assign(cur_rec->mlength.low, htot(low_word(SETUPMSIZE)));
+  assign(cur_rec->mlength.high, htot(high_word(SETUPMSIZE)));
 
   /* Process the kernel */
   flags = rdimage >= 0 ? 0 : FLAG_EOF;
@@ -203,19 +264,19 @@ static void do_kernel(void)
 	while (copyrec(KERNELNUM, SECTSIZE, kimage, kname) == SECTSIZE) ;
 	if (get_long(cur_rec->ilength) > BIG_SYSLSIZE) {
 		fprintf(stderr, "%s: Kernel image too large\n", progname);
-		exit(EXIT_FAILURE);
+		exit(EXIT_LINUX_KERNSIZE);
 	}
-	cur_rec->mlength.low  = htot(low_word(BIG_SYSMSIZE));
-	cur_rec->mlength.high = htot(high_word(BIG_SYSMSIZE));
+	assign(cur_rec->mlength.low, htot(low_word(BIG_SYSMSIZE)));
+	assign(cur_rec->mlength.high, htot(high_word(BIG_SYSMSIZE)));
   } else {
 	initrec(KERNELNUM, DEF_SYSSEG, flags, 0);
 	while (copyrec(KERNELNUM, SECTSIZE, kimage, kname) == SECTSIZE) ;
 	if (get_long(cur_rec->ilength) > SYSLSIZE) {
 		fprintf(stderr, "%s: Kernel image too large\n", progname);
-		exit(EXIT_FAILURE);
+		exit(EXIT_LINUX_KERNSIZE);
 	}
-	cur_rec->mlength.low  = htot(low_word(SYSMSIZE));
-	cur_rec->mlength.high = htot(high_word(SYSMSIZE));
+	assign(cur_rec->mlength.low, htot(low_word(SYSMSIZE)));
+	assign(cur_rec->mlength.high, htot(high_word(SYSMSIZE)));
   }
 
   /* Process the ramdisk image */
@@ -225,30 +286,32 @@ static void do_kernel(void)
 
 	if (imagestart < 0x100000L)
 		imagestart = 0x100000L;
-	initrec(RAMDISKNUM, 0, (rdmode == rd_eom ? FLAG_B1 : 0) | FLAG_EOF,
+	initrec(RAMDISKNUM, 0, (rdmode == RD_EOM ? FLAG_B1 : 0) | FLAG_EOF,
 					sizeof(cur_rec->vendor_data.rdflags));
-	if (rdmode == rd_auto)
+	if (rdmode == RD_AUTO)
 		rdlocation = imagestart;
-	else if (rdmode == rd_fixed)
+	else if (rdmode == RD_FIXED)
 		/* always align to 4kB page boundary */
 		rdlocation &= ~0xfff;
 	cur_rec->vendor_data.rdflags = (__u8)rdmode;
 	while (copyrec(RAMDISKNUM, SECTSIZE, rdimage, rdname) == SECTSIZE) ;
 	/* memory length has to be a multiple of 4kB */
-	l = (get_long(cur_rec->ilength) + 0xfff) & ~0xfff;
-	cur_rec->mlength.low  = htot(low_word(l));
-	cur_rec->mlength.high = htot(high_word(l));
-	l = rdmode == rd_eom ? get_long(cur_rec->mlength) : rdlocation;
-	cur_rec->address.low  = htot(low_word(l));
-	cur_rec->address.high = htot(high_word(l));
+	l = (get_long(cur_rec->ilength) + 0xfff) & (unsigned long)~0xfff;
+	assign(cur_rec->mlength.low, htot(low_word(l)));
+	assign(cur_rec->mlength.high, htot(high_word(l)));
+	l = rdmode == RD_EOM ? get_long(cur_rec->mlength) : rdlocation;
+	assign(cur_rec->address.low, htot(low_word(l)));
+	assign(cur_rec->address.high, htot(high_word(l)));
   }
 }
 
 
+
 /*
  * Remove duplicate and contradictory entries from the commandline
  */
-static void cleanup_cmdline(char *cmdline)
+static void cleanup_cmdline(cmdline)
+char *cmdline;
 {
   static char *entries[] = { " root=", " nfsroot=", " nfsaddrs=", NULL };
   char   **cur_entry;
@@ -262,7 +325,7 @@ static void cleanup_cmdline(char *cmdline)
    */
   if (strncmp(cmdline,"auto ",5)) {
 	fprintf(stderr, "%s: Internal error; command line mangled\n",progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_INTERNAL);
   }
 
   for (cur_entry = entries; *cur_entry; cur_entry++) {
@@ -287,24 +350,31 @@ again: /* Remove multiple occurrences of "ro" and "rw" */
 }
 
 
+
 /*
  * Dump the load record information to stderr
  */
-static void dump_header(struct load_header *lh)
+static void dump_header(lh)
+struct load_header *lh;
 {
-  static char *s_tags[] = { [BOOTLNUM]  "primary boot loader",
-			    [CMDLNUM]   "command line",
-			    [INITNUM]   "floppy boot sector",
-			    [SETUPNUM]  "kernel setup",
-			    [KERNELNUM] "kernel image",
-			    [RAMDISKNUM]"ramdisk image"};
+  static char *s_tags[] = { /* BOOTLNUM */   "primary boot loader",
+			    /* CMDLNUM */    "command line",
+			    /* INITNUM */    "floppy boot sector",
+			    /* SETUPNUM */   "kernel setup",
+			    /* KERNELNUM */  "kernel image",
+			    /* RAMDISKNUM */ "ramdisk image"};
   static char *s_flags[]= { "absolute address", "after previous segment",
 			    "at end of memory", "before previos segment"};
-  static char *s_rd[]   = { [RD_AUTO]   "RD: auto positioning",
-			    [RD_EOM]    "RD: positioned by Boot-Rom",
-			    [RD_FIXED]  "RD: fixed memory location" };
+  static char *s_rd[]   = { /* RD_AUTO */  "RD: auto positioning",
+			    /* RD_EOM */   "RD: positioned by Boot-Rom",
+			    /* RD_FIXED */ "RD: fixed memory location" };
   struct load_record *lr;
-  int    i, num = 0;
+  char *vendstr = NULL;
+  int i, num = 0;
+
+  i = (lh->hlength >> 2) & 0x3c;
+  vendstr = nbmalloc(i + 2);
+  memcpy(vendstr, lh->dummy, i);
 
   fprintf(stderr,"\n"
 	  "Load record information:\n"
@@ -321,15 +391,15 @@ static void dump_header(struct load_header *lh)
 	  (unsigned long)lh->hflags1 +
 		((unsigned long)lh->hflags2 << 8) +
 		((unsigned long)lh->hflags3 << 16),
-	  ttoh(lh->locn.segment), ttoh(lh->locn.offset),
-	  ttoh(lh->execute.segment), ttoh(lh->execute.offset),
-	  lh->dummy);
+	  ttoh(getval(lh->locn.segment)), ttoh(getval(lh->locn.offset)),
+	  ttoh(getval(lh->execute.segment)), ttoh(getval(lh->execute.offset)),
+	  vendstr);
 
   i  = ((lh->hlength >> 2) & 0x3c) + ((lh->hlength << 2) & 0x3c);
   lr = (struct load_record *)&(((__u8 *)lh)[i]);
 
   for (;;) {
-    fprintf(stderr,
+  	fprintf(stderr,
 	    "Record #%d:\n"
 	    "  Length of header: %d bytes (standard) + %d bytes (vendor)\n"
 	    "  Vendor tag:       0x%02X (%s)\n"
@@ -356,146 +426,135 @@ static void dump_header(struct load_header *lh)
 	    lr->vendor_data.rdflags < 3 ? s_rd[lr->vendor_data.rdflags] :
 	    "illegal" : "unknown" : "none");
 
-    if (lr->rflags & FLAG_EOF)
-      break;
+	if (lr->rflags & FLAG_EOF)
+		break;
 
-    i  = ((lr->rlength >> 2) & 0x3c) + ((lr->rlength << 2) & 0x3c);
-    lr = (struct load_record *)&(((__u8 *)lr)[i]);
+	i  = ((lr->rlength >> 2) & 0x3c) + ((lr->rlength << 2) & 0x3c);
+	lr = (struct load_record *)&(((__u8 *)lr)[i]);
   }
+  free(vendstr);
 }
 
 
+
 /*
- * Print the usage
+ * Read system database
  */
-static void usage(void)
+static void getdb(name)
+char *name;
 {
-  fprintf(stderr,
-	"Usage: %s [-a <append>] [-d rom|ram|kernel|<dir>]\n"
-	"             [-i rom|kernel|<client:server:gateway:netmask:hostname>]\n"
-	"             [-l auto|eom|<addr>] [-r <ramdisk>] [-s <max cmdline size]\n"
-	"             [-x] [[-k] <kernel>] [[-o] <outfile>]\n"
-	"       %s [-h]\n"
-	"       %s [-v]\n",
-	  progname, progname, progname);
-  exit(EXIT_FAILURE);
+  struct sectdef sect;
+  char *namebuf;
+  size_t len;
+
+  /* Read one entry from database file */
+  len = strlen(name) + 11;
+  namebuf = nbmalloc(len);
+  sprintf(namebuf, "%s:mknbi-linux", name);
+  sect.name = namebuf;
+  sect.params = dbparams;
+  sect.startsect = NULL;
+  sect.endsect = NULL;
+  readdb(&sect);
+
+  /* Check that parameters are correct */
+  if (cmdl_msize < CMDLMSIZE) {
+	fprintf(stderr, "%s: command line size must be >= %d bytes in section <%s>\n",
+						progname, CMDLMSIZE, namebuf);
+	exit(EXIT_DB);
+  }
+  if (kname == NULL) {
+	fprintf(stderr, "%s: Need kernel image file name in section <%s>\n",
+							progname, namebuf);
+	exit(EXIT_DB);
+  }
+  if (outname == NULL) {
+	fprintf(stderr, "%s: Need output file name in section <%s>\n",
+							progname, namebuf);
+	exit(EXIT_DB);
+  }
+  free(namebuf);
 }
 
 
+
 /*
  * Main program
  */
-void main(int argc, char **argv)
+void main(argc, argv)
+int argc;
+char **argv;
 {
-  char *cmdline = NULL;		/* Command line to pass to the kernel */
-  char *append = NULL;		/* String to append to end of command line */
-  char *addrs = NULL;		/* String containing various addresses */
+  size_t len;
+  char *cmdline;
   char *cp, *ip;
-  int cmdl_msize = CMDLMSIZE;
   int vendor_size;
-  int i, len;
-
-  /* Determine my own name for error output */
-  if ((cp = strrchr(argv[0], '/')) == NULL)
-	progname = argv[0];
-  else
-	progname = ++cp;
+  int i;
 
   /* Initialize option argments */
-  strcpy(nfsdir, DFLT_DIR);
-  strcpy(rootdev, DFLT_DEV);
-  strcpy(kname, DFLT_IMAGE);
-  strcpy(outname, "");
-  strcpy(rdname, "");
+  copystr(&nfsdir, DFLT_DIR);
+  copystr(&rootdev, DFLT_DEV);
+  copystr(&kname, DFLT_IMAGE);
 
-  /* Parse options */
-  opterr = 0;
-  while ((i = getopt(argc, argv, "a:d:hi:k:l:o:r:s:vx")) != EOF)
-	switch (i) {
-		case 'a': if ((append = strdup(optarg)) == NULL) {
-		          	perror(progname);
-		          	exit(EXIT_FAILURE);
-		          }
-		          break;
-		case 'd': strncpy(nfsdir, optarg, MAXPATHLEN-1);
-		          nfsdir[MAXPATHLEN-1] = '\0';
-		          break;
-		case 'i': if ((addrs = strdup(optarg)) == NULL) {
-				perror(progname);
-				exit(EXIT_FAILURE);
-		          }
-		          break;
-		case 'k': strncpy(kname, optarg, MAXPATHLEN-1);
-		          kname[MAXPATHLEN-1] = '\0';
-		          break;
-		case 'l': if (!strcmp("auto",optarg))
-				rdmode = rd_auto;
-		          else if (!strcmp("eom",optarg))
-				rdmode = rd_eom;
-		          else {
-				rdmode = rd_fixed;
-				rdlocation = strtol(optarg, NULL, 0); }
-		          break;
-		case 'o': strncpy(outname, optarg, MAXPATHLEN-1);
-		          outname[MAXPATHLEN-1] = '\0';
-		          break;
-		case 'r': strncpy(rdname, optarg, MAXPATHLEN-1);
-		          rdname[MAXPATHLEN-1] = '\0';
-		          break;
-		case 's': if ((i = atoi(optarg)) < CMDLMSIZE) {
-				fprintf(stderr, "%s: command line size "
-				                "must be > %d bytes\n",
-				        progname, CMDLMSIZE-1);
-				exit(EXIT_FAILURE);
-		          }
-		          cmdl_msize = i;
-			  break;
-		case 'v': fprintf(stderr, VERSION"\n");
-		          exit(EXIT_SUCCESS);
-		case 'x': verbose++;
-		          break;
-		case 'h': /* fall thru */
-		default:  usage();
+  /* Parse options and read configuration file */
+  nbsetup(argc, argv, opts, NULL);
+  if (batchname != NULL)
+	getdb(batchname);
+  if (cmdl_msize < CMDLMSIZE) {
+	fprintf(stderr, "%s: command line size must be >= %d bytes\n",
+						        progname, CMDLMSIZE-1);
+	exit(EXIT_USAGE);
+  }
+  if (kname == NULL) {
+	fprintf(stderr, "%s: Need kernel image file name\n", progname);
+	exit(EXIT_USAGE);
+  }
+  if (outname == NULL) {
+	fprintf(stderr, "%s: Need output file name\n", progname);
+	exit(EXIT_USAGE);
+  }
+
+  /* Decode ramdisk loading mode */
+  if (rdmodebuf == NULL || !strcmp("auto", rdmodebuf))
+	rdmode = RD_AUTO;
+  else if (!strcmp("eom", rdmodebuf))
+	rdmode = RD_EOM;
+  else {
+	rdmode = RD_FIXED;
+	if ((rdlocation = strtol(rdmodebuf, &cp, 0)) < 0x100000L) {
+		fprintf(stderr, "%s: Ramdisk image load location must be above 1M\n",
+								progname);
+		exit(EXIT_LINUX_RDLOC);
+	}
+	if (*cp) {
+		fprintf(stderr, "%s: invalid ramdisk location\n", progname);
+		exit(EXIT_LINUX_RDLOC);
 	}
-
-  /* Parse additional arguments */
-  if (optind < argc) {
-	strncpy(kname, argv[optind++], MAXPATHLEN-1);
-	kname[MAXPATHLEN-1] = '\0';
   }
-  if (optind < argc) {
-	strncpy(outname, argv[optind++], MAXPATHLEN-1);
-	outname[MAXPATHLEN-1] = '\0';
-  }
-  if (optind != argc)
-	usage();
 
   /* Parse the IP address option */
+#ifdef HAVE_INET
   if (addrs != NULL && *addrs && strcmp(addrs, "rom") &&
       strcmp(addrs, "kernel")) {
 	struct hostent *hp;
 	char *buf, *bp;
 
-	/* Allocate memory for 5 address strings, each with max 15 chars */
-	if ((buf = malloc((MAX_ADDR_SIZE + 1) * 5)) == NULL) {
-		perror(progname);
-		exit(EXIT_FAILURE);
-	}
+	/* Allocate memory for 6 address strings, each with max 15 chars */
 	i = 0;
 	ip = addrs;
-	bp = buf;
+	bp = buf = (char *)nbmalloc((MAX_ADDR_SIZE + 1) * 6);
 	while (ip != NULL && *ip) {
 		if ((cp = strchr(ip, ':')) != NULL) *cp++ = '\0';
 		if (strlen(ip) > 0) {
 			if ((hp = gethostbyname(ip)) == NULL) {
 				fprintf(stderr, "%s: invalid hostname %s\n",
 					progname, ip);
-				exit(EXIT_FAILURE);
+				exit(EXIT_HOSTNAME);
 			}
 			if (hp->h_length != sizeof(struct in_addr)) {
 				fprintf(stderr, "%s: invalid host address "
 					"type\n", progname);
-				exit(EXIT_FAILURE);
+				exit(EXIT_HOSTADDR);
 			}
 			/* Result from ntoa is smaller than MAX_ADDR_SIZE */
 			strcpy(bp, inet_ntoa(*(struct in_addr *) hp->h_addr));
@@ -508,87 +567,99 @@ void main(int argc, char **argv)
 	}
 	for (; i < 4; i++)
 		*bp++ = ':';
-	/* Finally copy host name */
+	/* Finally copy host and network card name */
 	if (ip != NULL) {
+		if ((cp = strchr(ip, ':')) != NULL) *cp++ = '\0';
 		if ((i = strlen(ip)) > MAX_ADDR_SIZE)
 			i = MAX_ADDR_SIZE;
 		strncpy(bp, ip, i);
 		bp += i;
+		if (cp != NULL) {
+			i = strlen(cp);
+			if (strncmp(cp, "eth", 3) || i != 4) {
+				fprintf(stderr, "%s: Invalid ethernet device "
+						"specification\n", progname);
+				exit(EXIT_LINUX_INVETH);
+			}
+			*bp++ = ':';
+			strncpy(bp, cp, i);
+			bp += i;
+		}
 	}
 	*bp = '\0';
 	free(addrs);
 	addrs = buf;
   }
-  if (addrs == NULL)
-	addrs = DFLT_ADDRS;
-
-  /* Verify validity of ramdisk load information */
-  if (rdmode == rd_fixed && rdlocation < 0x100000L) {
-	fprintf(stderr, "%s: Ramdisk image load location must be above 1M\n",
-								progname);
-	exit(EXIT_FAILURE);
+#else
+  if (addrs != NULL && *addrs && strcmp(addrs, "rom") &&
+      strcmp(addrs, "kernel")) {
+	fprintf(stderr, "%s: No INET support for -i option\n", progname);
+	exit(EXIT_INET);
   }
+#endif
+  if (addrs == NULL)
+	copystr(&addrs, DFLT_ADDRS);
 
   /* Determine root device for kernel */
   if (!strcmp(nfsdir, "ram")) {
 	if (strlen(rdname) == 0) {
 		fprintf(stderr, "%s: Boot from ramdisk but no image specified\n",
 								progname);
-		exit(EXIT_FAILURE);
+		exit(EXIT_LINUX_NORD);
 	}
+	if (rootdev != NULL)
+		free(rootdev);
+	rootdev = (char *)nbmalloc(strlen(nfsdir) + 12);
 	sprintf(rootdev, "root=/dev/%s", nfsdir);
-	strcpy(nfsdir, DFLT_DIR);
+	copystr(&nfsdir, DFLT_DIR);
   } else if (!strncmp(nfsdir, "/dev/", 5)) {
+	if (rootdev != NULL)
+		free(rootdev);
+	rootdev = (char *)nbmalloc(strlen(nfsdir) + 7);
 	sprintf(rootdev, "root=%s", nfsdir);
-	strcpy(nfsdir, DFLT_DIR);
+	copystr(&nfsdir, DFLT_DIR);
   }
 
   /* Open the input and output files */
-  if ((kimage = open(kname, O_RDONLY)) < 0) {
+  if ((kimage = open(kname, O_RDONLY | O_BINARY)) < 0) {
 	perror(kname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_KERNOPEN);
   }
-  if (strlen(outname) == 0 || !strcmp(outname, "-")) {
-	outfile = STDOUT_FILENO;
-	strcpy(outname, "stdout");
-  } else if ((outfile = creat(outname, 0664)) < 0) {
+  if ((outfile = creat(outname, 0644)) < 0) {
 	perror(outname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_IMGCREATE);
   }
-  if (*rdname && (rdimage = open(rdname, O_RDONLY)) < 0) {
+  if (rdname != NULL && (rdimage = open(rdname, O_RDONLY | O_BINARY)) < 0) {
 	perror(rdname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_RDOPEN);
   }
   if (verbose > 0) {
 	fprintf(stderr, "Kernel image file name  = %s\n", kname);
 	fprintf(stderr, "Output file name        = %s\n", outname);
-	if (*rdname)
+	if (rdname != NULL)
 		fprintf(stderr, "Ramdisk image file name = %s\n", rdname);
   }
 
   /* Construct command line to pass to the kernel */
   len = strlen(DFLT_CMDL) + strlen(rootdev) + 2;
-  if (append != NULL)
+  if (append != NULL && *append)
 	len += strlen(append) + 1;
-  if (*nfsdir)
+  if (nfsdir != NULL && *nfsdir)
 	len += strlen(NFS_ROOT) + strlen(nfsdir) + 1;
-  if (*addrs)
+  if (addrs != NULL && *addrs)
 	len += strlen(NFS_ADDRS) + strlen(addrs) + 1;
   if (len > CMDLLSIZE) {
 	fprintf(stderr, "%s: Command line too long\n", progname);
-	exit(EXIT_FAILURE);
-  }
-  if ((cmdline = malloc(len + 2)) == NULL) {
-	perror(progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_CMDLSIZE);
   }
+  cmdline = (char *)nbmalloc(len + 2);
 
   sprintf(cmdline, "%s %s", DFLT_CMDL, rootdev);
-  if (*nfsdir) {
+  if (nfsdir != NULL && *nfsdir) {
 	cp = cmdline + strlen(cmdline);
 	sprintf(cp, " %s%s", NFS_ROOT, nfsdir);
   }
-  if (*addrs) {
+  if (addrs != NULL && *addrs) {
 	cp = cmdline + strlen(cmdline);
 	sprintf(cp, " %s%s", NFS_ADDRS, addrs);
   }
@@ -604,22 +675,19 @@ void main(int argc, char **argv)
 
   /* Initialize the boot header */
   vendor_size = (sizeof(VENDOR_ID) / sizeof(__u32) + 1) * sizeof(__u32);
-  memset(&header, sizeof(header), 0);
-  header.magic.low       = htot(low_word(HEADER_MAGIC));
-  header.magic.high      = htot(high_word(HEADER_MAGIC));
-  header.hlength         = (__u8)(((int)&header.dummy - (int)&header)
+  memset(&header, 0, sizeof(header));
+  assign(header.magic.low,       htot(low_word(HEADER_MAGIC)));
+  assign(header.magic.high,      htot(high_word(HEADER_MAGIC)));
+  assign(header.locn.segment,    htot(DEF_HEADERSEG));
+  assign(header.locn.offset,     htot(0));
+  assign(header.execute.segment, htot(DEF_BOOTLSEG));
+  assign(header.execute.offset,  htot(0));
+  assign(header.bootsig,         htot(BOOT_SIGNATURE));
+  header.hlength         = (__u8)(((int)header.dummy - (int)&header)
                            / sizeof(__u32)) & 0x0f;
   header.hlength        |= (__u8)((vendor_size/sizeof(__u32)) << 4) & 0xf0;
-  header.locn.segment    = htot(DEF_HEADERSEG);
-  header.locn.offset     = 0;
-  header.execute.segment = htot(DEF_BOOTLSEG);
-  header.execute.offset  = 0;
-  header.bootsig         = htot(BOOT_SIGNATURE);
-  memcpy(&header.dummy, VENDOR_ID, sizeof(VENDOR_ID));
-  if (write(outfile, &header, sizeof(header)) < 0) {
-	perror(outname);
-	exit(EXIT_FAILURE);
-  }
+  memcpy(header.dummy, VENDOR_ID, sizeof(VENDOR_ID));
+  (void)nbwrite((unsigned char *)&header, sizeof(header), outfile);
 
   /* Initialize pointer to first load record */
   cur_rec = (struct load_record *)&(header.dummy[vendor_size]);
@@ -627,18 +695,18 @@ void main(int argc, char **argv)
   /* Process the boot loader record */
   if (first_data_size > BOOTLLSIZE) {
 	fprintf(stderr, "%s: Boot loader too large\n", progname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_LINUX_BOOTLSIZE);
   }
   initrec(BOOTLNUM, DEF_BOOTLSEG, 0, 0);
-  putrec(BOOTLNUM, (char *) &first_data, first_data_size);
-  cur_rec->mlength.low  = htot(low_word(BOOTLMSIZE));
-  cur_rec->mlength.high = htot(high_word(BOOTLMSIZE));
+  putrec(BOOTLNUM, (char *)first_data, first_data_size);
+  assign(cur_rec->mlength.low, htot(low_word(BOOTLMSIZE)));
+  assign(cur_rec->mlength.high, htot(high_word(BOOTLMSIZE)));
 
   /* Process the command line */
   initrec(CMDLNUM, DEF_CMDLSEG, 0, 0);
   putrec(CMDLNUM, cmdline, strlen(cmdline) + 1);
-  cur_rec->mlength.low  = htot(low_word(cmdl_msize));
-  cur_rec->mlength.high = htot(high_word(cmdl_msize));
+  assign(cur_rec->mlength.low, htot(low_word(cmdl_msize)));
+  assign(cur_rec->mlength.high, htot(high_word(cmdl_msize)));
 
   /* Process the kernel image */
   do_kernel();
@@ -646,12 +714,9 @@ void main(int argc, char **argv)
   /* After writing out all these stuff, finally update the boot header */
   if (lseek(outfile, 0, 0) != 0) {
 	perror(outname);
-	exit(EXIT_FAILURE);
-  }
-  if (write(outfile, &header, sizeof(header)) < 0) {
-	perror(outname);
-	exit(EXIT_FAILURE);
+	exit(EXIT_SEEK);
   }
+  (void)nbwrite((unsigned char *)&header, sizeof(header), outfile);
 
   /* If user asked for detailed output, parse the header and output all of */
   /* the load record information */
